import cv2
import imutils
from math import floor, ceil


def resize(image, width=500, height=500):
    """ Resizes and image, keeping the original aspect ratio
    :param image: the image to resize
    :param width: the max width of the new image
    :param height: the max height of the new image
    :return: the image resized to a max of 500x500 """
    image = imutils.resize(image, width=width, height=height)
    return image


def match_size(image_1, image_2, width=500, height=500):
    """ Resizes both images to the same size, padding when necessary
    :param image_1: the first image
    :param image_2: the second image
    :param width: the max width of the new images
    :param height: the max height of the new images
    :return: both images, resized to the same size """
    # Resize the data to a more manageable size
    image_1 = imutils.resize(image_1, width=width, height=height)
    image_2 = imutils.resize(image_2, width=width, height=height)

    # Pad the data to have matching dimensions
    height_1, width_1 = image_1.shape[:2]
    height_2, width_2 = image_2.shape[:2]
    vertical = height_1 - height_2
    horizontal = width_1 - width_2

    top = floor(vertical / 2)
    bottom = ceil(vertical / 2)
    left = floor(horizontal / 2)
    right = ceil(horizontal / 2)
    if vertical < 0:
        image_1 = cv2.copyMakeBorder(image_1, borderType=cv2.BORDER_CONSTANT,
                                     top=abs(top), bottom=abs(bottom), left=0, right=0)
    else:
        image_2 = cv2.copyMakeBorder(image_2, borderType=cv2.BORDER_CONSTANT,
                                     top=top, bottom=bottom, left=0, right=0)
    if horizontal < 0:
        image_1 = cv2.copyMakeBorder(image_1, borderType=cv2.BORDER_CONSTANT,
                                     top=0, bottom=0, left=abs(left), right=abs(right))
    else:
        image_2 = cv2.copyMakeBorder(image_2, borderType=cv2.BORDER_CONSTANT,
                                     top=0, bottom=0, left=left, right=right)

    return image_1, image_2



import cv2
import numpy as np


def ecc_alignment(image_1, image_2, iterations=1000, term_eps=1e-5):
    """ Takes two data, performs a perspective transformation to foreground orientation
    Returns second image as aligned to the first
    :param image_1: the image to which the second image should be aligned
    :param image_2: the image to align to the first image
    :param iterations: number of iterations
    :param term_eps: threshold of increment in correlation coefficient between iterations
    :return the aligned second image """
    rows, cols = image_1.shape[:2]

    image_1_gray = cv2.cvtColor(image_1, cv2.COLOR_BGR2GRAY)
    image_2_gray = cv2.cvtColor(image_2, cv2.COLOR_BGR2GRAY)

    warp_mode = cv2.MOTION_HOMOGRAPHY
    warp_matrix = np.eye(3, 3, dtype=np.float32)

    # Define termination criteria
    criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, iterations, term_eps)

    # Run the ECC algorithm. The results are stored in warp_matrix.
    cc, warp_matrix = cv2.findTransformECC(image_1_gray, image_2_gray, warp_matrix, warp_mode, criteria)

    # Warps image based on homography matrix from ECC algorithm
    image_2_aligned = cv2.warpPerspective(image_2, warp_matrix, (cols, rows),
                                          flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP)

    return image_2_aligned


def correspondence_warp(image_1, points_1, image_2, points_2):
    """ Warps the second image into the first image, based on the correspondences
    :param image_1: the first image
    :param points_1: ordered coordinates of the landmarks in the first image
    :param image_2: the second image
    :param points_2: ordered coordinates of the landmarks in the second image
    :return: the second image warped into the first image """
    for (x, y) in points_1:
        cv2.circle(image_1, (x, y), 3, 0, -1)
    for (x, y) in points_2:
        cv2.circle(image_2, (x, y), 3, 0, -1)

    # mat_warp = cv2.getPerspectiveTransform(points_2, points_1)
    mat_warp, status = cv2.findHomography(points_2, points_1)
    warped = cv2.warpPerspective(image_2, mat_warp, (image_1.shape[1], image_1.shape[0]))

    return warped



import cv2
import numpy as np
from sklearn.decomposition import PCA
from .segment import threshold_segment


def detect_moles(image, matching, min_area=10, max_area=100, pca_comp=0):
    """ Draws a circle around all moles in the image
    :param image: an image with moles to detect
    :param matching:
    :param min_area:
    :param max_area:
    :param pca_comp:
    :return the image with keypoints circled """
    params = cv2.SimpleBlobDetector_Params()

    # Filters
    params.filterByArea = True
    params.minArea = min_area
    params.maxArea = max_area
    params.filterByCircularity = False
    params.minCircularity = 0.1
    params.filterByConvexity = False
    params.minConvexity = 0.5
    params.filterByInertia = False
    params.minInertiaRatio = 0.01

    # Detect keypoints
    detector = cv2.SimpleBlobDetector_create(params)
    keypoints = detector.detect(image)

    # Compute descriptors
    descriptors = []
    comps = None
    if pca_comp > 0:
        orb = cv2.ORB_create()
        keypoints, descriptors = orb.compute(image, keypoints)

        pca = PCA(n_components=pca_comp)
        comps = pca.fit_transform(descriptors)

    # Use position as part of the descriptor
    if matching == 'NN':
        pad = 30 - pca_comp
        reduced = np.zeros((len(keypoints), pad))
        if comps:
            reduced = np.hstack((comps, reduced))

        descriptors = [np.array(keypoint.pt) for keypoint in keypoints]
        descriptors = np.hstack((descriptors, reduced)).astype(np.float32)

    return image, keypoints, descriptors


def detect_contours(image):
    """ Finds contours around skin.
    :param image: an image
    :return contour-overlay image and contour """
    _, skin_mask = threshold_segment(image)

    # Do contour detection on skin mask
    _, contours, hierarchy = cv2.findContours(skin_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    skin_contours = []
    # Draw the contour on the source image
    for i, c in enumerate(contours):
        # grab only contours of a certain size
        area = cv2.contourArea(c)
        if area > 6000:
            skin_contours.append(c)
            # specify precision as 1/1000 of total perimeter
            # peri = cv2.arcLength(c, True)
            # contours[i] = cv2.approxPolyDP(c, 0.001 * peri, True)
            # overlay the contour on the original image
            cv2.drawContours(image, contours, i, (0, 255, 0), 3)

    return image, skin_contours


def contour_overlay(base_image, overlay_image):
    """ Finds contours of base image and overlays them on the target image
    to help align data (for camera), returns overlayed image
    :param base_image: the base image
    :param overlay_image: the image on which to overlay contours
    :return the target image with the contours from the base image overlayed"""
    _, contours = detect_contours(base_image)

    for i, contour in enumerate(contours):
        cv2.drawContours(overlay_image, contours, i, (0, 0, 255), 3)

    return overlay_image



import cv2
import numpy as np


def image_stats(image):
    """ Splits an image into channels, and computes mean and std on each channel.
    :param image: an image to be split
    :return stats with the following format:
    (first.mean, first.std, second.mean, second.std, third.mean, third.std) """
    l, a, b = cv2.split(image)

    l_noblack = []
    a_noblack = []
    b_noblack = []

    for i, row in enumerate(l):
        for j, pixel in enumerate(row):
            if l[i][j] != 0:
                l_noblack.append(l[i][j])
                a_noblack.append(a[i][j])
                b_noblack.append(b[i][j])

    l_noblack = np.array(l_noblack)
    a_noblack = np.array(a_noblack)
    b_noblack = np.array(b_noblack)
    return l_noblack.mean(), l_noblack.std(), a_noblack.mean(), a_noblack.std(), b_noblack.mean(), b_noblack.std()


def clahe_equalize(image):
    """ Takes a color image, performs CLAHE (Contrast Limited Adaptive Histogram Equalization).
    :param image: image in BGR colorspace
    :return equalized image in BGR colorspace """

    # read in image and convert to HSV
    image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

    # split into channels
    h, s, v = cv2.split(image)

    # create CLAHE object
    clahe = cv2.createCLAHE(clipLimit=1.5, tileGridSize=(8, 8))
    equ_v = clahe.apply(v)

    # merge and convert back to BGR
    equ_image = cv2.merge((h, s, equ_v))
    equ_image = cv2.cvtColor(equ_image, cv2.COLOR_HSV2BGR)

    return equ_image


def transfer_color(source, target):
    """Takes the colormap of source image and maps it on to target image.
    For normalizing color between two different data.
    :param source: The source image
    :param target: The target image
    :return target image with source image color"""

    source = cv2.cvtColor(source, cv2.COLOR_BGR2LAB).astype("float32")
    target = cv2.cvtColor(target, cv2.COLOR_BGR2LAB).astype("float32")

    # color statistics for source and target image
    l_mean_src, l_std_src, a_mean_src, a_std_src, b_mean_src, b_std_src = image_stats(source)
    l_mean_tar, l_std_tar, a_mean_tar, a_std_tar, b_mean_tar, b_std_tar = image_stats(target)

    # subtract means from target image
    l, a, b = cv2.split(target)
    l -= l_mean_tar
    a -= a_mean_tar
    b -= b_mean_tar

    # scale by the standard deviation
    l = (l_std_tar / l_std_src) * l
    a = (a_std_tar / a_std_src) * a
    b = (b_std_tar / b_std_src) * b

    # add in the source mean
    l += l_mean_src
    a += a_mean_src
    b += b_mean_src

    # clip the pixel intensities to [0, 255] if they fall outside this range
    l = np.clip(l, 0, 255)
    a = np.clip(a, 0, 255)
    b = np.clip(b, 0, 255)

    # merge channels together and convert back to RGB color space
    # make sure to use 8-bit unsigned int
    transfer = cv2.merge([l, a, b])
    transfer_image = cv2.cvtColor(transfer.astype("uint8"), cv2.COLOR_LAB2BGR)

    return transfer_image



import cv2
import numpy as np


def enhance_edges(image, action='edge_enhance', kernel=None):
    """ Applies an edge filter to an image, accentuating edges
    :param image: the image to be filtered
    :param action: the desired action (sharpen, exc_sharpen, edge_enhance)
    :param kernel: an override kernel to apply to the image
    :return the image with edges enhanced """
    if not kernel:
        if action == 'sharpen':
            kernel = np.array([[-1, -1, -1],
                               [-1, +9, -1],
                               [-1, -1, -1]])
        elif action == 'exc_sharpen':
            kernel = np.array([[+1, +1, +1],
                               [+1, -7, +1],
                               [+1, +1, +1]])
        elif action == 'edge_enhance':
            kernel = np.array([[-1, -1, -1, -1, -1],
                               [-1, +2, +2, +2, -1],
                               [-1, +2, +8, +2, -1],
                               [-1, +2, +2, +2, -1],
                               [-1, -1, -1, -1, -1]]) / 8.0
        else:
            raise ValueError(action + ' not implemented')

    edge_image = cv2.filter2D(image, -1, kernel=kernel)
    return edge_image



import cv2
from skimage import io

from .color import clahe_equalize, transfer_color
from .filter import enhance_edges
from .preprocess import resize, match_size
from .register import ecc_alignment
from .segment import grabcut_segment
from .detect import detect_moles


def process(image_1_path, image_2_path=None, match='BF', matching='NN'):
    # -------------------------------------------------------- #
    # -------------------- Image Pipeline -------------------- #
    # -------------------------------------------------------- #
    image_2 = None
    match_image = None

    # First image arrives
    image_1 = io.imread(image_1_path)
    image_1 = resize(image_1)
    he_image_1 = clahe_equalize(image_1)

    # cv2.RECURS_FILTER
    # cv2.NORMCONV_FILTER
    image_1 = cv2.edgePreservingFilter(he_image_1, sigma_s=15, sigma_r=0.15)
    image_1 = enhance_edges(image_1)
    image_1, _ = grabcut_segment(image_1)

    if image_2_path:
        image_2 = io.imread(image_2_path)
        image_1, image_2 = match_size(image_1, image_2)
        he_image_1, _ = match_size(he_image_1, image_2)

        # Second image arrives
        image_2 = clahe_equalize(image_2)
        image_2 = transfer_color(he_image_1, image_2)
        image_2 = cv2.edgePreservingFilter(image_2, sigma_s=15, sigma_r=0.15)
        image_2 = enhance_edges(image_2)
        image_2, _ = grabcut_segment(image_2)
        image_2 = ecc_alignment(image_1, image_2)

        # Detect the moles
        image_1, keypoints_1, des_1 = detect_moles(image_1, matching)
        image_2, keypoints_2, des_2 = detect_moles(image_2, matching)

        # Match desciptors
        if match == 'BF':
            if matching == 'NN':
                matcher = cv2.BFMatcher(cv2.NORM_L2, crossCheck=True)
            else:
                matcher = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)
        else:
            if matching == 'NN':
                index_params = dict(algorithm=0, trees=5)
            else:
                index_params = dict(algorithm=6,
                                    table_number=12, key_size=20, multi_probe_level=2)
            search_params = dict(checks=50)  # or pass empty dictionary
            matcher = cv2.FlannBasedMatcher(index_params, search_params)

        # Find matches
        matches = matcher.match(des_1, des_2)
        matches = sorted(matches, key=lambda x: x.distance)
        match_image = cv2.drawMatches(image_1, keypoints_1, image_2, keypoints_2, matches, None)

    return image_1, image_2, match_image



from setuptools import setup

setup(
    name='omnimole',

    # Versions should comply with PEP440.
    version='1.0.9',

    description='An image processing toolbox for skin lesion classification',
    long_description='This package is for the image processing code used in the Omnium app. '
                     'The purpose of the app is to detect malignant skin lesions in hopes of '
                     'early diagnosis of skin cancer and melanoma.',
    url='https://github.com/omniumllc/image-processing',
    author='Omnium LLC',
    license='MIT',

    classifiers=[
        # Alpha, Beta, or Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        # 'Topic :: Image Processing :: Mole Classification',
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='omnium mole lesion image processing',
    packages=['omnimole'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed.
    install_requires=['imutils',
                      'numpy',
                      'opencv-python',
                      'PyMaxflow',
                      'scikit-image',
                      'scikit-learn'],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'omnimole=omnimole:main',
        ],
    },
)



import cv2
import maxflow
import numpy as np
from math import exp, pow


# Skin pixel values for YCrCb
# Y = 0-255
# Cr = 130-180
# Cb = 75-135

def grabcut_segment(image, rect=None):
    """ Segments the image, assuming that the outer border is representative of the
    background (possible rect would be (25, 0, cols - 50, rows))
    :param image: the original image
    :param rect: the bounding box which contains the foreground, or None for custom
    :return: the image with only the skin """
    rows, cols = image.shape[:2]

    mask_image = None
    mask = np.zeros((rows, cols), np.uint8)
    bg_model = np.zeros((1, 65), np.float64)
    fg_model = np.zeros((1, 65), np.float64)

    if rect:
        cv2.grabCut(image, mask, rect, bg_model, fg_model, 5, cv2.GC_INIT_WITH_RECT)
    else:
        r = rows / 2.
        c = cols / 2.
        d = r if r > c else c

        for i in range(rows):
            for j in range(cols):
                dist = np.sqrt((i - r) ** 2 + (j - c) ** 2)
                if dist < (1. / 3.) * d:
                    mask[i, j] = 1
                elif dist < (2. / 3.) * d:
                    mask[i, j] = 3
                elif dist < d:
                    mask[i, j] = 2

        mask_image = mask * 80
        cv2.grabCut(image, mask, None, bg_model, fg_model, 5, cv2.GC_INIT_WITH_MASK)

    mask = np.where((mask == 0) | (mask == 2), 0, 1).astype(np.uint8)
    grabcut = image * mask[:, :, np.newaxis]

    return grabcut, mask_image


def mrf_segment(image, cost_image, mu_1=130, mu_2=155, mu_3=105,
                sigma_1=100, sigma_2=25, sigma_3=30, pairwise=100):
    """ Segments an image as a Markov random field
    :param image: the original image
    :param cost_image: image on which to base the costs
    :param mu_1: mean of the first channel
    :param mu_2: mean of the second channel
    :param mu_3: mean of the third channel
    :param sigma_1: std of the first channel
    :param sigma_2: std of the second channel
    :param sigma_3: std of the third channel
    :param pairwise: the pairwise cost for the MRF
    :return: the segmented image """
    costs = [[gaussian_kernel_3(pixel, mu_1, sigma_1, mu_2, sigma_2, mu_3, sigma_3)
              for pixel in row] for row in cost_image]
    costs = np.array(costs)

    # Create graph
    g = maxflow.Graph[float]()
    nodes = g.add_grid_nodes(image.shape[:2])
    g.add_grid_edges(nodes, pairwise)
    g.add_grid_tedges(nodes, 100 - costs, costs)

    # Compute maxflow
    g.maxflow()

    # Mask the original image
    segments = g.get_grid_segments(nodes)
    labels = np.array(segments, dtype=np.uint8)
    mrf = cv2.bitwise_and(image, image, mask=labels)

    return mrf


def threshold_segment(image, space='ycrcb'):
    """ Takes an image applies a mask in the YCrCb colorspace to segment only skin.
    SOURCES:
    - Face Segmentation Using Skin-Color Map in Videophone Applications by Chai and Ngan
    - Skin Detection: A Step-by-Step Example Using Python and OpenCV by Adrian Rosebrock
    :param image: the image to segment
    :param space: the color space in which to segment the image
    :param lower: the lower threshold for skin in the color space
    :param upper: the higher threshold for skin in the color space
    :return filtered skin image in BGR color space and black and white mask """
    m, n = image.shape[:2]
    skin_mask = np.zeros((m, n), dtype=np.uint8)

    if space == 'hsv':
        lower = np.array([0, 45, 80], dtype=np.uint8)
        upper = np.array([20, 255, 255], dtype=np.uint8)
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        return threshold_skin(image, hsv, lower, upper)

    elif space == 'rgb':
        for i in range(0, m):
            for j in range(0, n):
                r, g, b = image[i][j]
                if r > 95 and g > 40 and b > 20 and r > g and r > b \
                        and max(r, g, b) - min(r, g, b) > 15 and abs(r - g) > 15:
                    skin_mask[i][j] = 255
        return mask_skin(image, skin_mask)

    elif space == 'ycrcb':
        lower = np.array([0, 130, 80], dtype=np.uint8)
        upper = np.array([255, 180, 130], dtype=np.uint8)
        ycrcb = cv2.cvtColor(image, cv2.COLOR_BGR2YCR_CB)
        return threshold_skin(image, ycrcb, lower, upper)

    elif space == 'yuv':
        converted = cv2.cvtColor(image, cv2.COLOR_BGR2YUV)
        for i in range(0, m):
            for j in range(0, n):
                r, g, b = image[i][j]
                y, u, v = converted[i][j]

                if 80 < u < 130 and 136 < v < 200 \
                        and r > 80 and g > 30 and b > 15 and abs(r - g) > 15:
                    skin_mask[i][j] = 255
        return mask_skin(image, skin_mask)

    else:
        raise ValueError(space + ' not implemented')


# -------------------------------------------------------- #
# -------------------- Helper Methods -------------------- #
# -------------------------------------------------------- #
def gaussian_kernel_2(pixel, mu_1, sigma_1, mu_2, sigma_2):
    d = exp(-(
        pow(mu_1 - pixel[0], 2) / (2 * pow(sigma_1, 2)) +
        pow(mu_2 - pixel[2], 2) / (2 * pow(sigma_2, 2))))
    return d * 100


def gaussian_kernel_3(pixel, mu_1, sigma_1, mu_2, sigma_2, mu_3, sigma_3):
    n = 1  # No normalization
    # n = 1 / (pow(2 * pi, 3 / 2) * y_sigma * cr_sigma * cb_sigma)

    d = exp(-(
        pow(mu_1 - pixel[0], 2) / (2 * pow(sigma_1, 2)) +
        pow(mu_2 - pixel[1], 2) / (2 * pow(sigma_2, 2)) +
        pow(mu_3 - pixel[2], 2) / (2 * pow(sigma_3, 2))))
    return n * d * 100


def threshold_moles(image, method='adaptive'):
    """ Performs adaptive or otsu thresholding on an image
    :param image: the original image
    :param method: 'adaptive' or 'otsu' thresholding
    :return: the thresholded image """
    gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    if method == 'adaptive':
        thresh = cv2.adaptiveThreshold(gray_image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                       cv2.THRESH_BINARY, 50, 50)
    elif method == 'otsu':
        _, thresh = cv2.threshold(gray_image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    else:
        raise ValueError(method + ' not implemented')

    image = cv2.bitwise_and(image, image, mask=thresh)
    return image


def threshold_skin(image, converted, lower, upper):
    skin_mask = cv2.inRange(converted, lower, upper)
    return mask_skin(image, skin_mask)


def mask_skin(image, skin_mask):
    # Apply erosions and dilations to mask
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (11, 11))
    skin_mask = cv2.erode(skin_mask, kernel, iterations=2)
    skin_mask = cv2.dilate(skin_mask, kernel, iterations=2)

    # Blur and apply mask
    skin_mask = cv2.GaussianBlur(skin_mask, (3, 3), 0)
    skin_image = cv2.bitwise_and(image, image, mask=skin_mask)

    return skin_image, skin_mask


def skinmask_overlay(base_image, overlay_image, alpha=0.5):
    """ Takes two data, takes the skinmask of the first image and overlays it on the
    seconds image (for camera) returns the overlayed image
    :param base_image: the base image
    :param overlay_image: the image to overlay
    :param alpha: the amount of transparency
    :return the overlayed image on the base image """
    skin_base, _ = threshold_segment(base_image)
    cv2.addWeighted(skin_base, alpha, overlay_image, 1 - alpha, 0, overlay_image)

    return overlay_image
